/*
 * Copyright 2003-2007 Sun Microsystems, Inc.  All Rights Reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa Clara,
 * CA 95054 USA or visit www.sun.com if you need additional information or
 * have any questions.
 */

#include "sunfontids.h"
#include "sunt2kscaler.h"
#include "GeneralPath.h"
#include <stdio.h>

extern "C" {

static void getGlyphGeneralPath(JNIEnv* env, jobject font2D, jlong pScalerContext, 
                                jint glyphCode, jfloat xpos, jfloat ypos,
                                GeneralPath* gp) {

    if (glyphCode >= INVISIBLE_GLYPHS) {
      return;
    }

    T2KScalerContext *context = (T2KScalerContext*)pScalerContext;
    T2KScalerInfo *scalerInfo = context->scalerInfo;
    T2K *t2k = scalerInfo->t2k;
    UInt32 renderFlags = ((context->t2kFlags | T2K_RETURN_OUTLINES)
			  & ~T2K_GRID_FIT);
    jboolean isQuadPath = scalerInfo->pathType == QUADPATHTYPE;

    int errCode;

    if (isNullScaler(scalerInfo) || context == theNullScalerContext) {
	return;
    }
    errCode = setupT2KContext(env, font2D, scalerInfo, context, false);
    if (errCode) {
        freeScalerInfoAfterError(env, context);
        return;
    }

    T2K_RenderGlyph(t2k, glyphCode, 0, 0, context->greyLevel, renderFlags,
		    &errCode);
    if (errCode) {
        t2kIfDebugMessage(errCode, "T2K_RenderGlyph failed", errCode);
        freeScalerInfoAfterError(env, context);
    }

    addGlyphToGeneralPath(*t2k->glyph, *gp, xpos, ypos, isQuadPath);

    T2K_PurgeMemory(t2k, 1, &errCode);
    if (errCode) {
        t2kIfDebugMessage(errCode, "T2K_PurgeMemory failed", errCode);
        freeScalerInfoAfterError(env, context);
    }
}

JNIEXPORT jobject JNICALL
Java_sun_font_FileFont_getGlyphOutline
    (JNIEnv *env, jobject font2D, jlong pScalerContext,
    jint glyphCode, jfloat xpos, jfloat ypos) {

    GeneralPath gp;
    getGlyphGeneralPath(env, font2D, pScalerContext, glyphCode, xpos, ypos, &gp);
    return gp.getShape(env);
}

JNIEXPORT jobject JNICALL
Java_sun_font_FileFont_getGlyphOutlineBounds
    (JNIEnv *env, jobject font2D, jlong pScalerContext,	int glyphCode) {

    GeneralPath gp;
    getGlyphGeneralPath(env, font2D, pScalerContext, glyphCode, 0.0f, 0.0f, &gp);
    return gp.getBounds(env);
}

JNIEXPORT jobject JNICALL
Java_sun_font_FileFont_getGlyphVectorOutline
    (JNIEnv *env, jobject font2D, jintArray glyphArray, int numGlyphs,
    jlong pScalerContext, jfloat xpos, jfloat ypos) {

    int i;
    T2KScalerContext *context = (T2KScalerContext*)pScalerContext;
    T2KScalerInfo *scalerInfo = context->scalerInfo;
    T2K *t2k = scalerInfo->t2k;
    GeneralPath generalPath;

    if (isNullScaler(scalerInfo) || context == theNullScalerContext) {
	return generalPath.getShape(env);
    }

    UInt32 renderFlags = ((context->t2kFlags | T2K_RETURN_OUTLINES)
			  & ~T2K_GRID_FIT);
    jint *glyphs = (jint*)malloc(sizeof(jint)*numGlyphs);
    jboolean isQuadPath = scalerInfo->pathType == QUADPATHTYPE;

    int errCode;

    errCode = setupT2KContext(env, font2D, scalerInfo, context, false);
    if (errCode) {
        freeScalerInfoAfterError(env, context);
	return generalPath.getShape(env);
    }

    env->GetIntArrayRegion(glyphArray, 0, numGlyphs, glyphs);
    for (i=0; i<numGlyphs;i++) {
	if (glyphs[i] >= INVISIBLE_GLYPHS) {
	    continue;
	}
	/* NB LCD modes have a T2K context which requests scaled bitmaps.
	 * Here we rely on T2K to interpret that to mean to scale only
	 * the returned bitmap image, not the returned outline.
	 */
	T2K_RenderGlyph(t2k, glyphs[i], 0, 0, context->greyLevel, renderFlags,
			&errCode);
        if (errCode) {
            t2kIfDebugMessage(errCode, "T2K_RenderGlyph failed", errCode);
            freeScalerInfoAfterError(env, context);
        }

	addGlyphToGeneralPath(*t2k->glyph, generalPath,
			      xpos, ypos, isQuadPath);

	T2K_PurgeMemory(t2k, 1, &errCode);
        if (errCode) {
            t2kIfDebugMessage(errCode, "T2K_PurgeMemory failed", errCode);
            freeScalerInfoAfterError(env, context);
        }
    }

    free(glyphs);
    return generalPath.getShape(env);
}

} /* End extern "C" */
