/*
 * Copyright 2003-2006 Sun Microsystems, Inc.  All Rights Reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa Clara,
 * CA 95054 USA or visit www.sun.com if you need additional information or
 * have any questions.
 */

#ifndef HEADLESS

#include <stdlib.h>
#include <jlong.h>

#include "OGLMaskBlit.h"
#include "OGLRenderQueue.h"
#include "OGLSurfaceData.h"
#include "GraphicsPrimitiveMgr.h"
#include "IntArgb.h"
#include "IntRgb.h"
#include "IntBgr.h"

#define MASK_LENGTH (OGLC_BLIT_TILE_SIZE * OGLC_BLIT_TILE_SIZE)

extern unsigned char mul8table[256][256];
extern OGLPixelFormat PixelFormats[];
static GLuint *rgbamask = NULL;

/**
 * This implementation of MaskBlit first combines the source system memory
 * tile with the corresponding alpha mask and stores the result in an
 * intermediate IntRgbaPre buffer.  That buffer is then copied into a cached
 * RGBA texture tile and then rendered to the destination surface.
 *
 * Note that currently there are only inner loops defined for IntArgb,
 * IntRgb, and IntBgr, as those are the most commonly used formats for this
 * operation.
 *
 * REMIND: this method assumes that rgbamask and the cached blit texture tile
 *         have the same dimensions, and that the incoming maskArray is
 *         of equal or lesser dimensions; these are rather fragile
 *         assumptions, and should be cleaned up at some point...
 */
void
OGLMaskBlit_MaskBlit(JNIEnv *env, OGLContext *oglc,
                     jlong pSrcOps, jint srctype,
                     jint srcx, jint srcy,
                     jint dstx, jint dsty,
                     jint width, jint height,
                     jint maskoff, jint maskscan, jint masklen,
                     unsigned char *pMask)
{
    SurfaceDataOps *srcOps = (SurfaceDataOps *)jlong_to_ptr(pSrcOps);
    SurfaceDataRasInfo srcInfo;

    J2dTraceLn(J2D_TRACE_INFO, "OGLMaskBlit_MaskBlit");

    if (width <= 0 || height <= 0) {
        J2dTraceLn(J2D_TRACE_WARNING,
                   "OGLMaskBlit_MaskBlit: invalid dimensions");
        return;
    }

    RETURN_IF_NULL(srcOps);
    RETURN_IF_NULL(pMask);
    RETURN_IF_NULL(oglc);
    CHECK_PREVIOUS_OP(GL_TEXTURE_2D);

    if (rgbamask == NULL) {
        rgbamask = (GLuint *)malloc(sizeof(GLuint) * MASK_LENGTH);
        if (rgbamask == NULL) {
            J2dRlsTraceLn(J2D_TRACE_ERROR,
                "OGLMaskBlit_MaskBlit: could not allocate rgbamask");
            return;
        }
    }

    if (oglc->blitTextureID == 0) {
        if (!OGLContext_InitBlitTileTexture(oglc)) {
            J2dRlsTraceLn(J2D_TRACE_ERROR,
                "OGLMaskBlit_MaskBlit: could not init blit tile");
            return;
        }
    }

    srcInfo.bounds.x1 = srcx;
    srcInfo.bounds.y1 = srcy;
    srcInfo.bounds.x2 = srcx + width;
    srcInfo.bounds.y2 = srcy + height;

    if (srcOps->Lock(env, srcOps, &srcInfo, SD_LOCK_READ) != SD_SUCCESS) {
        J2dTraceLn(J2D_TRACE_WARNING,
                   "OGLMaskBlit_MaskBlit: could not acquire lock");
	return;
    }

    if (srcInfo.bounds.x2 > srcInfo.bounds.x1 &&
	srcInfo.bounds.y2 > srcInfo.bounds.y1)
    {
        srcOps->GetRasInfo(env, srcOps, &srcInfo);
	if (srcInfo.rasBase) {
            jfloat tx1, ty1, tx2, ty2;
            GLubyte ea;
            jint h;
            jint srcScanStride = srcInfo.scanStride;
            jint srcPixelStride = srcInfo.pixelStride;
            void *pSrcBase = PtrCoord(srcInfo.rasBase,
                                      srcInfo.bounds.x1, srcInfo.pixelStride,
                                      srcInfo.bounds.y1, srcInfo.scanStride);
            jint *pSrc = (jint *)pSrcBase;
            unsigned char *pMaskBase = pMask;
            jint newscan;
            GLuint *rgbavals;

            if (masklen > MASK_LENGTH) {
                // REMIND: this approach is seriously flawed if the mask
                //         length is ever greater than MASK_LENGTH (won't fit
                //         into the cached mask tile)...
                J2dRlsTraceLn(J2D_TRACE_ERROR,
                    "OGLMaskBlit_MaskBlit: mask array too large");
                SurfaceData_InvokeRelease(env, srcOps, &srcInfo);
                SurfaceData_InvokeUnlock(env, srcOps, &srcInfo);
                return;
            }

            width = srcInfo.bounds.x2 - srcInfo.bounds.x1;
            height = srcInfo.bounds.y2 - srcInfo.bounds.y1;
            maskoff += ((srcInfo.bounds.y1 - srcy) * maskscan +
                        (srcInfo.bounds.x1 - srcx));
            srcScanStride -= width * srcPixelStride;
            newscan = maskscan - width;
            pMask += maskoff;
            rgbavals = rgbamask + maskoff;
            ea = (GLubyte)(oglc->extraAlpha * 255);
            h = height;

            J2dTraceLn4(J2D_TRACE_VERBOSE,
                        "  sx=%d sy=%d w=%d h=%d",
                        srcInfo.bounds.x1, srcInfo.bounds.y1, width, height);
            J2dTraceLn2(J2D_TRACE_VERBOSE,
                        "  dx=%d dy=%d",
                        dstx, dsty);
            J2dTraceLn3(J2D_TRACE_VERBOSE,
                        "  maskoff=%d maskscan=%d newscan=%d",
                        maskoff, maskscan, newscan);
            J2dTraceLn2(J2D_TRACE_VERBOSE,
                        "  pixstride=%d scanstride=%d",
                        srcPixelStride, srcScanStride);

            // apply alpha values from mask to the source tile, and store
            // results in an intermediate "IntRgbaPre" buffer (there are
            // separate inner loops for the three most common source formats)
            switch (srctype) {
            case sun_java2d_opengl_OGLSurfaceData_PF_INT_ARGB:
                do {
                    jint w = width;
                    do {
                        jint pathA = *pMask++;
                        if (!pathA) {
                            rgbavals[0] = 0;
                        } else {
                            jint cr, cg, cb, ca;
                            GLubyte r, g, b, a;
                            LoadIntArgbTo4ByteArgb(pSrc, c, 0, ca, cr, cg, cb);
                            a = MUL8(MUL8(ca, pathA), ea);
                            r = MUL8(cr, a);
                            g = MUL8(cg, a);
                            b = MUL8(cb, a);
                            rgbavals[0] = (r << 24) | (g << 16) | (b << 8) | a;
                        }
                        pSrc = PtrAddBytes(pSrc, srcPixelStride);
                        rgbavals++;
                    } while (--w > 0);
                    pSrc = PtrAddBytes(pSrc, srcScanStride);
                    pMask = PtrAddBytes(pMask, newscan);
                    rgbavals = PtrAddBytes(rgbavals, newscan * 4);
                } while (--h > 0);
                break;

            case sun_java2d_opengl_OGLSurfaceData_PF_INT_RGB:
                do {
                    jint w = width;
                    do {
                        jint pathA = *pMask++;
                        if (!pathA) {
                            rgbavals[0] = 0;
                        } else {
                            jint cr, cg, cb;
                            GLubyte r, g, b, a;
                            LoadIntRgbTo3ByteRgb(pSrc, c, 0, cr, cg, cb);
                            a = MUL8(pathA, ea);
                            r = MUL8(cr, a);
                            g = MUL8(cg, a);
                            b = MUL8(cb, a);
                            rgbavals[0] = (r << 24) | (g << 16) | (b << 8) | a;
                        }
                        pSrc = PtrAddBytes(pSrc, srcPixelStride);
                        rgbavals++;
                    } while (--w > 0);
                    pSrc = PtrAddBytes(pSrc, srcScanStride);
                    pMask = PtrAddBytes(pMask, newscan);
                    rgbavals = PtrAddBytes(rgbavals, newscan * 4);
                } while (--h > 0);
                break;

            case sun_java2d_opengl_OGLSurfaceData_PF_INT_BGR:
                do {
                    jint w = width;
                    do {
                        jint pathA = *pMask++;
                        if (!pathA) {
                            rgbavals[0] = 0;
                        } else {
                            jint cr, cg, cb;
                            GLubyte r, g, b, a;
                            LoadIntBgrTo3ByteRgb(pSrc, c, 0, cr, cg, cb);
                            a = MUL8(pathA, ea);
                            r = MUL8(cr, a);
                            g = MUL8(cg, a);
                            b = MUL8(cb, a);
                            rgbavals[0] = (r << 24) | (g << 16) | (b << 8) | a;
                        }
                        pSrc = PtrAddBytes(pSrc, srcPixelStride);
                        rgbavals++;
                    } while (--w > 0);
                    pSrc = PtrAddBytes(pSrc, srcScanStride);
                    pMask = PtrAddBytes(pMask, newscan);
                    rgbavals = PtrAddBytes(rgbavals, newscan * 4);
                } while (--h > 0);
                break;

            default:
                // should not get here, just no-op...
                break;
            }

            // set up texture parameters
            j2d_glBindTexture(GL_TEXTURE_2D, oglc->blitTextureID);
            OGLC_UPDATE_TEXTURE_FUNCTION(oglc, GL_REPLACE);
            j2d_glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER,
                                GL_NEAREST);
            j2d_glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
                                GL_NEAREST);

            j2d_glPixelStorei(GL_UNPACK_ROW_LENGTH, maskscan);
            j2d_glPixelStorei(GL_UNPACK_SKIP_ROWS, maskoff / maskscan);
            j2d_glPixelStorei(GL_UNPACK_SKIP_PIXELS, maskoff % maskscan);
            j2d_glPixelStorei(GL_UNPACK_ALIGNMENT, 4);

            // copy system memory IntRgbaPre surface into cached texture
            j2d_glTexSubImage2D(GL_TEXTURE_2D, 0,
                                0, 0, width, height,
                                GL_RGBA, GL_UNSIGNED_INT_8_8_8_8, rgbamask);

            tx1 = 0.0f;
            ty1 = 0.0f;
            tx2 = ((GLfloat)width) / OGLC_BLIT_TILE_SIZE;
            ty2 = ((GLfloat)height) / OGLC_BLIT_TILE_SIZE;

            // render cached texture to the OpenGL surface
            j2d_glBegin(GL_QUADS);
            j2d_glTexCoord2f(tx1, ty1);
            j2d_glVertex2i(dstx, dsty);
            j2d_glTexCoord2f(tx2, ty1);
            j2d_glVertex2i(dstx + width, dsty);
            j2d_glTexCoord2f(tx2, ty2);
            j2d_glVertex2i(dstx + width, dsty + height);
            j2d_glTexCoord2f(tx1, ty2);
            j2d_glVertex2i(dstx, dsty + height);
            j2d_glEnd();

            j2d_glPixelStorei(GL_UNPACK_ROW_LENGTH, 0);
            j2d_glPixelStorei(GL_UNPACK_SKIP_ROWS, 0);
            j2d_glPixelStorei(GL_UNPACK_SKIP_PIXELS, 0);
        }
	SurfaceData_InvokeRelease(env, srcOps, &srcInfo);
    }
    SurfaceData_InvokeUnlock(env, srcOps, &srcInfo);
}

#endif /* !HEADLESS */
