/*
 * Copyright 1996-2007 Sun Microsystems, Inc.  All Rights Reserved.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Sun designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Sun in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Sun Microsystems, Inc., 4150 Network Circle, Santa Clara,
 * CA 95054 USA or visit www.sun.com if you need additional information or
 * have any questions.
 */
package sun.awt.windows;

import java.awt.*;
import java.awt.peer.*;
import java.lang.ref.WeakReference;
import java.lang.reflect.Method;
import sun.awt.SunToolkit;
import sun.awt.Win32GraphicsDevice;
import sun.awt.PaintEventDispatcher;

class WCanvasPeer extends WComponentPeer implements CanvasPeer {

    private boolean eraseBackgroundDisabledOnThis;

    private static boolean eraseBackgroundDisabled;
    private static boolean eraseBackgroundOnResize;
    static {
        String prop;
        prop = (String)java.security.AccessController.doPrivileged(
            new sun.security.action.GetPropertyAction("sun.awt.noerasebackground"));
        eraseBackgroundDisabled = (prop != null
                              && prop.length() > 0
                              && prop.charAt(0) == 't');
        prop = (String)java.security.AccessController.doPrivileged(
            new sun.security.action.GetPropertyAction("sun.awt.erasebackgroundonresize"));
        eraseBackgroundOnResize = (prop != null
                              && prop.length() > 0
                              && prop.charAt(0) == 't');
    }

    Method resetGCMethod;

    // Toolkit & peer internals

    WCanvasPeer(Component target) {
        super(target);
    }

    /*
     * From the DisplayChangedListener interface.
     *
     * Overrides WComponentPeer version because Canvases can be created with
     * a non-defulat GraphicsConfiguration, which is no longer valid.
     * Up-called for other windows peer instances (WPanelPeer, WWindowPeer).
     */
    public void displayChanged() {
        clearLocalGC();
        resetTargetGC();
        super.displayChanged();
    }

    /*
     * Reset the graphicsConfiguration member of our target Component.
     * Component.resetGC() is a package-private method, so we have to call it
     * through reflection.
     */
    public void resetTargetGC() {
        synchronized (this) {
            if (resetGCMethod == null) {
                resetGCMethod = SunToolkit.getMethod(Component.class, "resetGC", null);
            }
        }
        try {
            resetGCMethod.invoke(target);
        } catch (Exception z) {
            z.printStackTrace();
        }
    }

    /*
     * Clears the peer's winGraphicsConfig member.
     * Overridden by WWindowPeer, which shouldn't have a null winGraphicsConfig.
     */
    void clearLocalGC() {
        winGraphicsConfig = null;
    }

    native void create(WComponentPeer parent);

    private native void setNativeBackgroundErase(boolean doErase);

    void initialize() {
        // Optimization: the default value in the native code is true, so we 
        // call setNativeBackgroundErase only when the value changes to false
        if (!PaintEventDispatcher.getPaintEventDispatcher().
                shouldDoNativeBackgroundErase((Component)target)) {
            setNativeBackgroundErase(false);
        }
        super.initialize();
	Color bg = ((Component)target).getBackground();
	if (bg != null) {
	    setBackground(bg);
	}
    }

    public void paint(Graphics g) {
	Dimension d = ((Component)target).getSize();
        if (g instanceof Graphics2D ||
	    g instanceof sun.awt.Graphics2Delegate) {
	    // background color is setup correctly, so just use clearRect
	    g.clearRect(0, 0, d.width, d.height);
	} else {
	    // emulate clearRect
	    g.setColor(((Component)target).getBackground());
	    g.fillRect(0, 0, d.width, d.height);
	    g.setColor(((Component)target).getForeground());
	}
	super.paint(g);
    }

    public void print(Graphics g) {
	Dimension d = ((Component)target).getSize();
        if (g instanceof Graphics2D ||
	    g instanceof sun.awt.Graphics2Delegate) {
	    // background color is setup correctly, so just use clearRect
	    g.clearRect(0, 0, d.width, d.height);
	} else {
	    // emulate clearRect
	    g.setColor(((Component)target).getBackground());
	    g.fillRect(0, 0, d.width, d.height);
	    g.setColor(((Component)target).getForeground());
	}
	super.print(g);
    }
    
    public boolean shouldClearRectBeforePaint() {
        return (eraseBackgroundDisabled == false) && (eraseBackgroundDisabledOnThis == false);
    }

    void disableBackgroundErase() {
        eraseBackgroundDisabledOnThis = true;
        disableNativeBackgroundErase();
    }

    private native void disableNativeBackgroundErase();

}
